###########################################################################################
# Training script for MACE
# Authors: Ilyes Batatia, Gregor Simm, David Kovacs
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import ast
import glob
import json
import logging
import os
from copy import deepcopy
from pathlib import Path
from typing import List, Optional

import torch.distributed
from macer.externals.e3nn_legacy.util import jit
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.optim import LBFGS
from torch.utils.data import ConcatDataset
from torch_ema import ExponentialMovingAverage

import macer.externals.mace_bundled as mace
from macer.externals.mace_bundled import data, tools
from macer.externals.mace_bundled.calculators.foundations_models import mace_mp, mace_mp_names, mace_off
from macer.externals.mace_bundled.cli.convert_cueq_e3nn import run as run_cueq_to_e3nn
from macer.externals.mace_bundled.cli.convert_e3nn_cueq import run as run_e3nn_to_cueq
from macer.externals.mace_bundled.cli.convert_e3nn_oeq import run as run_e3nn_to_oeq
from macer.externals.mace_bundled.cli.convert_oeq_e3nn import run as run_oeq_to_e3nn
from macer.externals.mace_bundled.cli.visualise_train import TrainingPlotter
from macer.externals.mace_bundled.data import KeySpecification, update_keyspec_from_kwargs
from macer.externals.mace_bundled.tools import torch_geometric
from macer.externals.mace_bundled.tools.distributed_tools import init_distributed
from macer.externals.mace_bundled.tools.model_script_utils import configure_model
from macer.externals.mace_bundled.tools.multihead_tools import (
    HeadConfig,
    apply_pseudolabels_to_pt_head_configs,
    assemble_replay_data,
    dict_head_to_dataclass,
    prepare_default_head,
    prepare_pt_head,
)
from macer.externals.mace_bundled.tools.run_train_utils import (
    combine_datasets,
    load_dataset_for_path,
    normalize_file_paths,
)
from macer.externals.mace_bundled.tools.scripts_utils import (
    LRScheduler,
    SubsetCollection,
    check_path_ase_read,
    convert_to_json_format,
    dict_to_array,
    extract_config_mace_model,
    get_atomic_energies,
    get_avg_num_neighbors,
    get_config_type_weights,
    get_dataset_from_xyz,
    get_files_with_suffix,
    get_loss_fn,
    get_optimizer,
    get_params_options,
    get_swa,
    print_git_commit,
    remove_pt_head,
    setup_wandb,
)
from macer.externals.mace_bundled.tools.tables_utils import create_error_table
from macer.externals.mace_bundled.tools.utils import AtomicNumberTable


def main() -> None:
    """
    This script runs the training/fine tuning for mace
    """
    args = tools.build_default_arg_parser().parse_args()
    run(args)


def run(args) -> None:
    """
    This script runs the training/fine tuning for mace
    """
    tag = tools.get_tag(name=args.name, seed=args.seed)
    args, input_log_messages = tools.check_args(args)

    # default keyspec to update using heads dictionary
    args.key_specification = KeySpecification()
    update_keyspec_from_kwargs(args.key_specification, vars(args))

    if args.device == "xpu":
        try:
            import intel_extension_for_pytorch as ipex
            import oneccl_bindings_for_pytorch as oneccl  # pylint: disable=unused-import
        except ImportError as e:
            raise ImportError(
                "Error: Intel extension for PyTorch not found, but XPU device was specified"
            ) from e
    rank, local_rank, world_size = init_distributed(args)

    # Setup
    tools.set_seeds(args.seed)
    tools.setup_logger(level=args.log_level, tag=tag, directory=args.log_dir, rank=rank)
    logging.info("===========VERIFYING SETTINGS===========")
    for message, loglevel in input_log_messages:
        logging.log(level=loglevel, msg=message)

    if args.distributed:
        if args.device == "cuda":
            torch.cuda.set_device(local_rank)
        elif args.device == "xpu":
            torch.xpu.set_device(local_rank)
        logging.info(f"Process group initialized: {torch.distributed.is_initialized()}")
        logging.info(f"Processes: {world_size}")

    try:
        logging.info(f"MACE version: {mace.__version__}")
    except AttributeError:
        logging.info("Cannot find MACE version, please install MACE via pip")
    logging.debug(f"Configuration: {args}")

    tools.set_default_dtype(args.default_dtype)
    device = tools.init_device(args.device)
    commit = print_git_commit()
    model_foundation: Optional[torch.nn.Module] = None
    foundation_model_avg_num_neighbors = 0
    # Filter out None from mace_mp_names to get valid model names
    valid_mace_mp_models = [name for name in mace_mp_names if name is not None]
    if args.foundation_model is not None:
        if args.foundation_model in valid_mace_mp_models:
            logging.info(
                f"Using foundation model mace-mp-0 {args.foundation_model} as initial checkpoint."
            )
            calc = mace_mp(
                model=args.foundation_model,
                device=args.device,
                default_dtype=args.default_dtype,
            )
            model_foundation = calc.models[0]
        elif args.foundation_model in ["small_off", "medium_off", "large_off"]:
            model_type = args.foundation_model.split("_")[0]
            logging.info(
                f"Using foundation model mace-off-2023 {model_type} as initial checkpoint. ASL license."
            )
            calc = mace_off(
                model=model_type,
                device=args.device,
                default_dtype=args.default_dtype,
            )
            model_foundation = calc.models[0]
        else:
            model_foundation = torch.load(
                args.foundation_model, map_location=args.device
            )
            logging.info(
                f"Using foundation model {args.foundation_model} as initial checkpoint."
            )
        args.r_max = model_foundation.r_max.item()
        foundation_model_avg_num_neighbors = model_foundation.interactions[
            0
        ].avg_num_neighbors
        if (
            args.foundation_model not in ["small", "medium", "large"]
            and args.pt_train_file is None
        ):
            logging.warning(
                "Using multiheads finetuning with a foundation model that is not a Materials Project model, need to provied a path to a pretraining file with --pt_train_file."
            )
            args.multiheads_finetuning = False
        if args.multiheads_finetuning:
            assert (
                args.E0s != "average"
            ), "average atomic energies cannot be used for multiheads finetuning"
            # check that the foundation model has a single head, if not, use the first head
            if not args.force_mh_ft_lr:
                logging.info(
                    "Multihead finetuning mode, setting learning rate to 0.0001 and EMA to True. To use a different learning rate, set --force_mh_ft_lr=True."
                )
                args.lr = 0.0001
                args.ema = True
                args.ema_decay = 0.99999
            logging.info(
                "Using multiheads finetuning mode, setting learning rate to 0.0001 and EMA to True"
            )
            if hasattr(model_foundation, "heads"):
                if len(model_foundation.heads) > 1:
                    logging.warning(
                        "Mutlihead finetuning with models with more than one head is not supported, using the first head as foundation head."
                    )
                    model_foundation = remove_pt_head(
                        model_foundation, args.foundation_head
                    )
    else:
        args.multiheads_finetuning = False

    if args.heads is not None:
        args.heads = ast.literal_eval(args.heads)
        for _, head_dict in args.heads.items():
            # priority is global args < head property_key values < head info_keys+arrays_keys
            head_keyspec = deepcopy(args.key_specification)
            update_keyspec_from_kwargs(head_keyspec, head_dict)
            head_keyspec.update(
                info_keys=head_dict.get("info_keys", {}),
                arrays_keys=head_dict.get("arrays_keys", {}),
            )
            head_dict["key_specification"] = head_keyspec
    else:
        args.heads = prepare_default_head(args)
    if args.multiheads_finetuning:
        pt_keyspec = (
            args.heads["pt_head"]["key_specification"]
            if "pt_head" in args.heads
            else deepcopy(args.key_specification)
        )
        args.heads["pt_head"] = prepare_pt_head(
            args, pt_keyspec, foundation_model_avg_num_neighbors
        )

    logging.info("===========LOADING INPUT DATA===========")
    heads = list(args.heads.keys())
    logging.info(f"Using heads: {heads}")
    logging.info("Using the key specifications to parse data:")
    for name, head_dict in args.heads.items():
        head_keyspec = head_dict["key_specification"]
        logging.info(f"{name}: {head_keyspec}")

    head_configs: List[HeadConfig] = []
    for head, head_args in args.heads.items():
        logging.info(f"=============    Processing head {head}     ===========")
        head_config = dict_head_to_dataclass(head_args, head, args)
        # don't apply user's --atomic_numbers to pt_head, that info needs to come
        # from the actual pt data
        if args.multiheads_finetuning and head_config.head_name == "pt_head":
            head_config.atomic_numbers = None

        # Handle train_file and valid_file - normalize to lists
        if hasattr(head_config, "train_file") and head_config.train_file is not None:
            head_config.train_file = normalize_file_paths(head_config.train_file)
        if hasattr(head_config, "valid_file") and head_config.valid_file is not None:
            head_config.valid_file = normalize_file_paths(head_config.valid_file)
        if hasattr(head_config, "test_file") and head_config.test_file is not None:
            head_config.test_file = normalize_file_paths(head_config.test_file)

        if (
            head_config.statistics_file is not None
            and head_config.head_name != "pt_head"
        ):
            with open(head_config.statistics_file, "r") as f:  # pylint: disable=W1514
                statistics = json.load(f)
            logging.info("Using statistics json file")
            head_config.atomic_numbers = statistics["atomic_numbers"]
            head_config.mean = statistics["mean"]
            head_config.std = statistics["std"]
            head_config.avg_num_neighbors = statistics["avg_num_neighbors"]
            head_config.compute_avg_num_neighbors = False
            if isinstance(statistics["atomic_energies"], str) and statistics[
                "atomic_energies"
            ].endswith(".json"):
                with open(statistics["atomic_energies"], "r", encoding="utf-8") as f:
                    atomic_energies = json.load(f)
                head_config.E0s = atomic_energies
                head_config.atomic_energies_dict = ast.literal_eval(atomic_energies)
            else:
                head_config.E0s = statistics["atomic_energies"]
                head_config.atomic_energies_dict = ast.literal_eval(
                    statistics["atomic_energies"]
                )
        if head_config.train_file in (["mp"], ["matpes_pbe"], ["matpes_r2scan"]):
            assert (
                head_config.head_name == "pt_head"
            ), "Only pt_head should use mp as train_file"
            logging.info(
                f"Using filtered Materials Project data for replay ({args.num_samples_pt}, {args.filter_type_pt}, {args.subselect_pt}). "
                "You can also construct a different subset using `fine_tuning_select.py` script."
            )
            collections = assemble_replay_data(
                head_config.train_file[0], args, head_config, tag
            )
            head_config.collections = collections
        elif any(check_path_ase_read(f) for f in head_config.train_file):
            train_files_ase_list = [
                f for f in head_config.train_file if check_path_ase_read(f)
            ]
            valid_files_ase_list = None
            test_files_ase_list = None
            if head_config.valid_file:
                valid_files_ase_list = [
                    f for f in head_config.valid_file if check_path_ase_read(f)
                ]
            if head_config.test_file:
                test_files_ase_list = [
                    f for f in head_config.test_file if check_path_ase_read(f)
                ]
            config_type_weights = get_config_type_weights(
                head_config.config_type_weights
            )
            collections, atomic_energies_dict = get_dataset_from_xyz(
                work_dir=args.work_dir,
                train_path=train_files_ase_list,
                valid_path=valid_files_ase_list,
                valid_fraction=head_config.valid_fraction,
                config_type_weights=config_type_weights,
                test_path=test_files_ase_list,
                seed=args.seed,
                key_specification=head_config.key_specification,
                head_name=head_config.head_name,
                keep_isolated_atoms=head_config.keep_isolated_atoms,
                no_data_ok=(
                    args.pseudolabel_replay
                    and args.multiheads_finetuning
                    and head_config.head_name == "pt_head"
                ),
            )
            head_config.collections = SubsetCollection(
                train=collections.train,
                valid=collections.valid,
                tests=collections.tests,
            )
            head_config.atomic_energies_dict = atomic_energies_dict
            logging.info(
                f"Total number of configurations: train={len(collections.train)}, valid={len(collections.valid)}, "
                f"tests=[{', '.join([name + ': ' + str(len(test_configs)) for name, test_configs in collections.tests])}],"
            )
        head_configs.append(head_config)

    if all(
        check_path_ase_read(head_config.train_file[0]) for head_config in head_configs
    ):
        size_collections_train = sum(
            len(head_config.collections.train) for head_config in head_configs
        )
        size_collections_valid = sum(
            len(head_config.collections.valid) for head_config in head_configs
        )
        if size_collections_train < args.batch_size:
            logging.error(
                f"Batch size ({args.batch_size}) is larger than the number of training data ({size_collections_train})"
            )
        if size_collections_valid < args.valid_batch_size:
            logging.warning(
                f"Validation batch size ({args.valid_batch_size}) is larger than the number of validation data ({size_collections_valid})"
            )

    if args.multiheads_finetuning:
        logging.info(
            "==================Using multiheads finetuning mode=================="
        )
        args.loss = "universal"

        all_ase_readable = all(
            all(check_path_ase_read(f) for f in head_config.train_file)
            for head_config in head_configs
        )
        head_config_pt = filter(lambda x: x.head_name == "pt_head", head_configs)
        head_config_pt = next(head_config_pt, None)
        assert head_config_pt is not None, "Pretraining head not found"
        if all_ase_readable:
            ratio_pt_ft = (
                size_collections_train - len(head_config_pt.collections.train)
            ) / len(head_config_pt.collections.train)
            if ratio_pt_ft < args.real_pt_data_ratio_threshold:
                logging.warning(
                    f"Ratio of the number of configurations in the training set and the in the pt_train_file is {ratio_pt_ft}, "
                    f"increasing the number of configurations in the fine-tuning heads by {int(args.real_pt_data_ratio_threshold / ratio_pt_ft)}"
                )
                for head_config in head_configs:
                    if head_config.head_name == "pt_head":
                        continue
                    head_config.collections.train += (
                        head_config.collections.train
                        * int(args.real_pt_data_ratio_threshold / ratio_pt_ft)
                    )
            logging.info(
                f"Total number of configurations in pretraining: train={len(head_config_pt.collections.train)}, valid={len(head_config_pt.collections.valid)}"
            )
        else:
            logging.debug(
                "Using LMDB/HDF5 datasets for pretraining or fine-tuning - skipping ratio check"
            )

    # Atomic number table
    # yapf: disable
    for head_config in head_configs:
        if head_config.atomic_numbers is None:
            assert all(check_path_ase_read(f) for f in head_config.train_file), "Must specify atomic_numbers when using .h5 or .aselmdb train_file input"
            z_table_head = tools.get_atomic_number_table_from_zs(
                z
                for configs in (head_config.collections.train, head_config.collections.valid)
                for config in configs
                for z in config.atomic_numbers
            )
            head_config.atomic_numbers = z_table_head.zs
            head_config.z_table = z_table_head
        else:
            if head_config.statistics_file is None:
                logging.info("Using atomic numbers from command line argument")
            else:
                logging.info("Using atomic numbers from statistics file")
            zs_list = ast.literal_eval(head_config.atomic_numbers)
            assert isinstance(zs_list, list)
            z_table_head = tools.AtomicNumberTable(zs_list)
            head_config.atomic_numbers = zs_list
            head_config.z_table = z_table_head
        # yapf: enable
    all_atomic_numbers = set()
    for head_config in head_configs:
        all_atomic_numbers.update(head_config.atomic_numbers)
    z_table = AtomicNumberTable(sorted(list(all_atomic_numbers)))
    if args.foundation_model_elements and model_foundation:
        z_table = AtomicNumberTable(sorted(model_foundation.atomic_numbers.tolist()))
    logging.info(f"Atomic Numbers used: {z_table.zs}")

    # Atomic energies
    atomic_energies_dict = {}
    for head_config in head_configs:
        if head_config.atomic_energies_dict is None or len(head_config.atomic_energies_dict) == 0:
            assert head_config.E0s is not None, "Atomic energies must be provided"
            if all(check_path_ase_read(f) for f in head_config.train_file) and head_config.E0s.lower() != "foundation":
                atomic_energies_dict[head_config.head_name] = get_atomic_energies(
                    head_config.E0s, head_config.collections.train, head_config.z_table
                )
            elif head_config.E0s.lower() == "foundation":
                assert args.foundation_model is not None
                z_table_foundation = AtomicNumberTable(
                    [int(z) for z in model_foundation.atomic_numbers]
                )
                foundation_atomic_energies = model_foundation.atomic_energies_fn.atomic_energies
                if foundation_atomic_energies.ndim > 1:
                    foundation_atomic_energies = foundation_atomic_energies.squeeze()
                    if foundation_atomic_energies.ndim == 2:
                        foundation_atomic_energies = foundation_atomic_energies[0]
                        logging.info("Foundation model has multiple heads, using the first head as foundation E0s.")
                atomic_energies_dict[head_config.head_name] = {
                    z: foundation_atomic_energies[
                        z_table_foundation.z_to_index(z)
                    ].item()
                    for z in z_table.zs
                }
            else:
                atomic_energies_dict[head_config.head_name] = get_atomic_energies(head_config.E0s, None, head_config.z_table)
        else:
            atomic_energies_dict[head_config.head_name] = head_config.atomic_energies_dict

    # Atomic energies for multiheads finetuning
    if args.multiheads_finetuning:
        assert (
            model_foundation is not None
        ), "Model foundation must be provided for multiheads finetuning"
        z_table_foundation = AtomicNumberTable(
            [int(z) for z in model_foundation.atomic_numbers]
        )
        foundation_atomic_energies = model_foundation.atomic_energies_fn.atomic_energies
        if foundation_atomic_energies.ndim > 1:
            foundation_atomic_energies = foundation_atomic_energies.squeeze()
            if foundation_atomic_energies.ndim == 2:
                foundation_atomic_energies = foundation_atomic_energies[0]
                logging.info("Foundation model has multiple heads, using the first head as foundation E0s.")
        atomic_energies_dict["pt_head"] = {
            z: foundation_atomic_energies[
                z_table_foundation.z_to_index(z)
            ].item()
            for z in z_table.zs
        }
    heads = sorted(heads, key=lambda x: -1000 if x == "pt_head" else 0)
    # Padding atomic energies if keeping all elements of the foundation model
    if args.foundation_model_elements and model_foundation:
        atomic_energies_dict_padded = {}
        for head_name, head_energies in atomic_energies_dict.items():
            energy_head_padded = {}
            for z in z_table.zs:
                energy_head_padded[z] = head_energies.get(z, 0.0)
            atomic_energies_dict_padded[head_name] = energy_head_padded
        atomic_energies_dict = atomic_energies_dict_padded

    if args.model == "AtomicDipolesMACE":
        atomic_energies = None
        dipole_only = True
        args.compute_dipole = True
        args.compute_energy = False
        args.compute_forces = False
        args.compute_virials = False
        args.compute_stress = False
        args.compute_polarizability = False
    elif args.model == "AtomicDielectricMACE":
        atomic_energies = None
        dipole_only = False
        args.compute_dipole = True
        args.compute_energy = False
        args.compute_forces = False
        args.compute_virials = False
        args.compute_stress = False
        args.compute_polarizability = True
    else:
        dipole_only = False
        if args.model == "EnergyDipolesMACE":
            args.compute_dipole = True
            args.compute_energy = True
            args.compute_forces = True
            args.compute_virials = False
            args.compute_stress = False
            args.compute_polarizability = False
        else:
            args.compute_energy = True
            args.compute_dipole = False
            args.compute_polarizability = False
        # atomic_energies: np.ndarray = np.array(
        #     [atomic_energies_dict[z] for z in z_table.zs]
        # )
        atomic_energies = dict_to_array(atomic_energies_dict, heads)
        for head_config in head_configs:
            try:
                logging.info(f"Atomic Energies used (z: eV) for head {head_config.head_name}: " + "{" + ", ".join([f"{z}: {atomic_energies_dict[head_config.head_name][z]}" for z in head_config.z_table.zs]) + "}")
            except KeyError as e:
                raise KeyError(f"Atomic number {e} not found in atomic_energies_dict for head {head_config.head_name}, add E0s for this atomic number") from e

    # Load datasets for each head, supporting multiple files per head
    valid_sets = {head: [] for head in heads}
    train_sets = {head: [] for head in heads}

    for head_config in head_configs:
        train_datasets = []

        logging.info(f"Processing datasets for head '{head_config.head_name}'")

        # Apply pseudolabels if this is the pt_head and pseudolabeling is enabled
        if args.pseudolabel_replay and args.multiheads_finetuning and head_config.head_name == "pt_head":
            logging.info("=============    Pseudolabeling for pt_head    ===========")
            if apply_pseudolabels_to_pt_head_configs(
                foundation_model=model_foundation,
                pt_head_config=head_config,
                r_max=args.r_max,
                device=device,
                batch_size=args.batch_size
            ):
                logging.info("Successfully applied pseudolabels to pt_head configurations")
            else:
                logging.warning("Pseudolabeling was not successful, continuing with original configurations")

        ase_files = [f for f in head_config.train_file if check_path_ase_read(f)]
        non_ase_files = [f for f in head_config.train_file if not check_path_ase_read(f)]

        if ase_files:
            dataset = load_dataset_for_path(
            file_path=ase_files,
            r_max=args.r_max,
            z_table=z_table,
            head_config=head_config,
            heads=heads,
            collection=head_config.collections.train,
            )
            train_datasets.append(dataset)
            logging.debug(f"Successfully loaded dataset from ASE files: {ase_files}")

        for file in non_ase_files:
            dataset = load_dataset_for_path(
            file_path=file,
            r_max=args.r_max,
            z_table=z_table,
            head_config=head_config,
            heads=heads,
            )
            train_datasets.append(dataset)
            logging.debug(f"Successfully loaded dataset from non-ASE file: {file}")

        if not train_datasets:
            raise ValueError(f"No valid training datasets found for head {head_config.head_name}")

        train_sets[head_config.head_name] = combine_datasets(train_datasets, head_config.head_name)

        if head_config.valid_file:
            valid_datasets = []

            valid_ase_files = [f for f in head_config.valid_file if check_path_ase_read(f)]
            valid_non_ase_files = [f for f in head_config.valid_file if not check_path_ase_read(f)]

            if valid_ase_files:
                valid_dataset = load_dataset_for_path(
                    file_path=valid_ase_files,
                    r_max=args.r_max,
                    z_table=z_table,
                    head_config=head_config,
                    heads=heads,
                    collection=head_config.collections.valid,
                )
                valid_datasets.append(valid_dataset)
                logging.debug(f"Successfully loaded validation dataset from ASE files: {valid_ase_files}")
            for valid_file in valid_non_ase_files:
                valid_dataset = load_dataset_for_path(
                file_path=valid_file,
                r_max=args.r_max,
                z_table=z_table,
                head_config=head_config,
                heads=heads,
            )
                valid_datasets.append(valid_dataset)
                logging.debug(f"Successfully loaded validation dataset from {valid_file}")

            # Combine validation datasets
            if valid_datasets:
                valid_sets[head_config.head_name] = combine_datasets(valid_datasets, f"{head_config.head_name}_valid")
                logging.info(f"Combined validation datasets for {head_config.head_name}")

        # If no valid file is provided but collection exist, use the validation set from the collection
        if head_config.valid_file is None and head_config.collections.valid:
            valid_sets[head_config.head_name] = [
                data.AtomicData.from_config(
                    config, z_table=z_table, cutoff=args.r_max, heads=heads
                )
                for config in head_config.collections.valid
            ]
        if not valid_sets[head_config.head_name]:
            raise ValueError(f"No valid datasets found for head {head_config.head_name}, please provide a valid_file or a valid_fraction")

        # Create data loader for this head
        if isinstance(train_sets[head_config.head_name], list):
            dataset_size = len(train_sets[head_config.head_name])
        else:
            dataset_size = len(train_sets[head_config.head_name])
        logging.info(f"Head '{head_config.head_name}' training dataset size: {dataset_size}")

        train_loader_head = torch_geometric.dataloader.DataLoader(
            dataset=train_sets[head_config.head_name],
            batch_size=args.batch_size,
            shuffle=True,
            drop_last=(not args.lbfgs),
            pin_memory=args.pin_memory,
            num_workers=args.num_workers,
            generator=torch.Generator().manual_seed(args.seed),
        )
        head_config.train_loader = train_loader_head

    # concatenate all the trainsets
    train_set = ConcatDataset([train_sets[head] for head in heads])
    train_sampler, valid_sampler = None, None
    if args.distributed:
        train_sampler = torch.utils.data.distributed.DistributedSampler(
            train_set,
            num_replicas=world_size,
            rank=rank,
            shuffle=True,
            drop_last=(not args.lbfgs),
            seed=args.seed,
        )
        valid_samplers = {}
        for head, valid_set in valid_sets.items():
            valid_sampler = torch.utils.data.distributed.DistributedSampler(
                valid_set,
                num_replicas=world_size,
                rank=rank,
                shuffle=True,
                drop_last=True,
                seed=args.seed,
            )
            valid_samplers[head] = valid_sampler

    train_loader = torch_geometric.dataloader.DataLoader(
        dataset=train_set,
        batch_size=args.batch_size,
        sampler=train_sampler,
        shuffle=(train_sampler is None),
        drop_last=(train_sampler is None and not args.lbfgs),
        pin_memory=args.pin_memory,
        num_workers=args.num_workers,
        generator=torch.Generator().manual_seed(args.seed),
    )

    valid_loaders = {heads[i]: None for i in range(len(heads))}
    if not isinstance(valid_sets, dict):
        valid_sets = {"Default": valid_sets}
    for head, valid_set in valid_sets.items():
        valid_loaders[head] = torch_geometric.dataloader.DataLoader(
            dataset=valid_set,
            batch_size=args.valid_batch_size,
            sampler=valid_samplers[head] if args.distributed else None,
            shuffle=False,
            drop_last=False,
            pin_memory=args.pin_memory,
            num_workers=args.num_workers,
            generator=torch.Generator().manual_seed(args.seed),
        )

    loss_fn = get_loss_fn(args, dipole_only, args.compute_dipole)
    args.avg_num_neighbors = get_avg_num_neighbors(head_configs, args, train_loader, device)

    # Model
    model, output_args = configure_model(args, train_loader, atomic_energies, model_foundation, heads, z_table, head_configs)
    model.to(device)

    logging.debug(model)
    logging.info(f"Total number of parameters: {tools.count_parameters(model)}")
    logging.info("")
    logging.info("===========OPTIMIZER INFORMATION===========")
    logging.info(f"Using {args.optimizer.upper()} as parameter optimizer")
    logging.info(f"Batch size: {args.batch_size}")
    if args.ema:
        logging.info(f"Using Exponential Moving Average with decay: {args.ema_decay}")
    logging.info(
        f"Number of gradient updates: {int(args.max_num_epochs*len(train_set)/args.batch_size)}"
    )
    logging.info(f"Learning rate: {args.lr}, weight decay: {args.weight_decay}")
    logging.info(loss_fn)

    # Cueq and OEQ conversion
    if args.enable_cueq and args.enable_oeq:
        logging.warning(
            "Both CUEQ and OEQ are enabled, using CUEQ for training. "
            "To use OEQ, disable CUEQ with --disable_cueq."
        )
        args.enable_oeq = False
    if args.enable_cueq and not args.only_cueq:
        logging.info("Converting model to CUEQ for accelerated training")
        assert model.__class__.__name__ in ["MACE", "ScaleShiftMACE", "MACELES"]
        model = run_e3nn_to_cueq(deepcopy(model), device=device)
    if args.enable_oeq:
        logging.info("Converting model to OEQ for accelerated training")
        assert model.__class__.__name__ in ["MACE", "ScaleShiftMACE", "MACELES"]
        model = run_e3nn_to_oeq(deepcopy(model), device=device)

    # Optimizer
    param_options = get_params_options(args, model)
    optimizer: torch.optim.Optimizer
    optimizer = get_optimizer(args, param_options)
    if args.device == "xpu":
        logging.info("Optimzing model and optimzier for XPU")
        model, optimizer = ipex.optimize(model, optimizer=optimizer)
    logger = tools.MetricsLogger(
        directory=args.results_dir, tag=tag + "_train"
    )  # pylint: disable=E1123

    lr_scheduler = LRScheduler(optimizer, args)

    swa: Optional[tools.SWAContainer] = None
    swas = [False]
    if args.swa:
        swa, swas = get_swa(args, model, optimizer, swas, dipole_only)

    checkpoint_handler = tools.CheckpointHandler(
        directory=args.checkpoints_dir,
        tag=tag,
        keep=args.keep_checkpoints,
        swa_start=args.start_swa,
    )

    start_epoch = 0
    restart_lbfgs = False
    opt_start_epoch = None
    if args.restart_latest:
        try:
            opt_start_epoch = checkpoint_handler.load_latest(
                state=tools.CheckpointState(model, optimizer, lr_scheduler),
                swa=True,
                device=device,
            )
        except Exception:  # pylint: disable=W0703
            try:
                opt_start_epoch = checkpoint_handler.load_latest(
                    state=tools.CheckpointState(model, optimizer, lr_scheduler),
                    swa=False,
                    device=device,
                )
            except Exception: # pylint: disable=W0703
                restart_lbfgs = True
        if opt_start_epoch is not None:
            start_epoch = opt_start_epoch

    ema: Optional[ExponentialMovingAverage] = None
    if args.ema:
        ema = ExponentialMovingAverage(model.parameters(), decay=args.ema_decay)
    else:
        for group in optimizer.param_groups:
            group["lr"] = args.lr

    if args.lbfgs:
        logging.info("Switching optimizer to LBFGS")
        optimizer = LBFGS(model.parameters(),
                          history_size=200,
                          max_iter=20,
                          line_search_fn="strong_wolfe")
        if restart_lbfgs:
            opt_start_epoch = checkpoint_handler.load_latest(
                state=tools.CheckpointState(model, optimizer, lr_scheduler),
                swa=False,
                device=device,
            )
            if opt_start_epoch is not None:
                start_epoch = opt_start_epoch

    if args.wandb:
        setup_wandb(args)
    if args.distributed:
        distributed_model = DDP(model, device_ids=[local_rank])
    else:
        distributed_model = None


    train_valid_data_loader = {}
    for head_config in head_configs:
        data_loader_name = "train_" + head_config.head_name
        train_valid_data_loader[data_loader_name] = head_config.train_loader
    for head, valid_loader in valid_loaders.items():
        data_load_name = "valid_" + head
        train_valid_data_loader[data_load_name] = valid_loader

    if args.plot and args.plot_frequency > 0:
        try:
            plotter = TrainingPlotter(
                results_dir=logger.path,
                heads=heads,
                table_type=args.error_table,
                train_valid_data=train_valid_data_loader,
                test_data={},
                output_args=output_args,
                device=device,
                plot_frequency=args.plot_frequency,
                distributed=args.distributed,
                swa_start=swa.start if swa else None,
                plot_interaction_e=args.plot_interaction_e
                )
        except Exception as e:  # pylint: disable=W0718
            logging.debug(f"Creating Plotter failed: {e}")
    else:
        plotter = None

    if args.dry_run:
        logging.info("DRY RUN mode enabled. Stopping now.")
        return

    if args.device == "xpu":
        try:
            model, optimizer = ipex.optimize(model, optimizer=optimizer)
        except ImportError as e:
            logging.error(
                "Intel Extension for PyTorch not found, but XPU device was specified. "
                "Please install it to use XPU device."
            )

    tools.train(
        model=model,
        loss_fn=loss_fn,
        train_loader=train_loader,
        valid_loaders=valid_loaders,
        optimizer=optimizer,
        lr_scheduler=lr_scheduler,
        checkpoint_handler=checkpoint_handler,
        eval_interval=args.eval_interval,
        start_epoch=start_epoch,
        max_num_epochs=args.max_num_epochs,
        logger=logger,
        patience=args.patience,
        save_all_checkpoints=args.save_all_checkpoints,
        output_args=output_args,
        device=device,
        swa=swa,
        ema=ema,
        max_grad_norm=args.clip_grad,
        log_errors=args.error_table,
        log_wandb=args.wandb,
        distributed=args.distributed,
        distributed_model=distributed_model,
        plotter=plotter,
        train_sampler=train_sampler,
        rank=rank,
    )

    logging.info("")
    logging.info("===========RESULTS===========")

    train_valid_data_loader = {}
    for head_config in head_configs:
        data_loader_name = "train_" + head_config.head_name
        train_valid_data_loader[data_loader_name] = head_config.train_loader
    for head, valid_loader in valid_loaders.items():
        data_load_name = "valid_" + head
        train_valid_data_loader[data_load_name] = valid_loader
    test_sets = {}
    stop_first_test = False
    test_data_loader = {}
    if all(
        head_config.test_file == head_configs[0].test_file
        for head_config in head_configs
    ) and head_configs[0].test_file is not None:
        stop_first_test = True
    if all(
        head_config.test_dir == head_configs[0].test_dir
        for head_config in head_configs
    ) and head_configs[0].test_dir is not None:
        stop_first_test = True
    for head_config in head_configs:
        if all(check_path_ase_read(f) for f in head_config.train_file):
            for name, subset in head_config.collections.tests:
                test_sets[name] = [
                    data.AtomicData.from_config(
                        config, z_table=z_table, cutoff=args.r_max, heads=heads
                    )
                    for config in subset
                ]
        if head_config.test_dir is not None:
            if not args.multi_processed_test:
                test_files = get_files_with_suffix(head_config.test_dir, "_test.h5")
                for test_file in test_files:
                    name = os.path.splitext(os.path.basename(test_file))[0]
                    test_sets[name] = data.HDF5Dataset(
                        test_file, r_max=args.r_max, z_table=z_table, heads=heads, head=head_config.head_name
                    )
            else:
                test_folders = glob(head_config.test_dir + "/*")
                for folder in test_folders:
                    name = os.path.splitext(os.path.basename(test_file))[0]
                    test_sets[name] = data.dataset_from_sharded_hdf5(
                        folder, r_max=args.r_max, z_table=z_table, heads=heads, head=head_config.head_name
                    )
        for test_name, test_set in test_sets.items():
            test_sampler = None
            if args.distributed:
                test_sampler = torch.utils.data.distributed.DistributedSampler(
                    test_set,
                    num_replicas=world_size,
                    rank=rank,
                    shuffle=True,
                    drop_last=True,
                    seed=args.seed,
                )
            try:
                drop_last = test_set.drop_last
            except AttributeError as e:  # pylint: disable=W0612
                drop_last = False
            test_loader = torch_geometric.dataloader.DataLoader(
                test_set,
                batch_size=args.valid_batch_size,
                shuffle=(test_sampler is None),
                drop_last=drop_last,
                num_workers=args.num_workers,
                pin_memory=args.pin_memory,
            )
            test_data_loader[test_name] = test_loader
        if stop_first_test:
            break

    for swa_eval in swas:
        epoch = checkpoint_handler.load_latest(
            state=tools.CheckpointState(model, optimizer, lr_scheduler),
            swa=swa_eval,
            device=device,
        )
        model.to(device)
        if args.distributed:
            distributed_model = DDP(model, device_ids=[local_rank])
        model_to_evaluate = model if not args.distributed else distributed_model
        if swa_eval:
            logging.info(f"Loaded Stage two model from epoch {epoch} for evaluation")
        else:
            logging.info(f"Loaded Stage one model from epoch {epoch} for evaluation")

        if rank == 0:
            # Save entire model
            if swa_eval:
                model_path = Path(args.checkpoints_dir) / (tag + "_stagetwo.model")
            else:
                model_path = Path(args.checkpoints_dir) / (tag + ".model")
            logging.info(f"Saving model to {model_path}")
            model_to_save = deepcopy(model)
            if args.enable_cueq and not args.only_cueq:
                logging.info("RUNING CUEQ TO E3NN")
                model_to_save = run_cueq_to_e3nn(deepcopy(model), device=device)
            if args.enable_oeq:
                logging.info("RUNING OEQ TO E3NN")
                model_to_save = run_oeq_to_e3nn(deepcopy(model), device=device)
            if args.save_cpu:
                model_to_save = model_to_save.to("cpu")
            torch.save(model_to_save, model_path)
            extra_files = {
                "commit.txt": commit.encode("utf-8") if commit is not None else b"",
                "config.yaml": json.dumps(
                    convert_to_json_format(extract_config_mace_model(model))
                ),
            }
            if swa_eval:
                torch.save(
                    model_to_save, Path(args.model_dir) / (args.name + "_stagetwo.model")
                )
                try:
                    path_complied = Path(args.model_dir) / (
                        args.name + "_stagetwo_compiled.model"
                    )
                    logging.info(f"Compiling model, saving metadata {path_complied}")
                    model_compiled = jit.compile(deepcopy(model_to_save))
                    torch.jit.save(
                        model_compiled,
                        path_complied,
                        _extra_files=extra_files,
                    )
                except Exception as e:  # pylint: disable=W0718
                    pass
            else:
                torch.save(model_to_save, Path(args.model_dir) / (args.name + ".model"))
                try:
                    path_complied = Path(args.model_dir) / (
                        args.name + "_compiled.model"
                    )
                    logging.info(f"Compiling model, saving metadata to {path_complied}")
                    model_compiled = jit.compile(deepcopy(model_to_save))
                    torch.jit.save(
                        model_compiled,
                        path_complied,
                        _extra_files=extra_files,
                    )
                except Exception as e:  # pylint: disable=W0718
                    pass

        logging.info("Computing metrics for training, validation, and test sets")
        for param in model.parameters():
            param.requires_grad = False
        skip_heads = args.skip_evaluate_heads.split(",") if args.skip_evaluate_heads else []
        if skip_heads:
            logging.info(f"Skipping evaluation for heads: {skip_heads}")
        table_train_valid = create_error_table(
            table_type=args.error_table,
            all_data_loaders=train_valid_data_loader,
            model=model_to_evaluate,
            loss_fn=loss_fn,
            output_args=output_args,
            log_wandb=args.wandb,
            device=device,
            distributed=args.distributed,
            skip_heads=skip_heads,
        )
        logging.info("Error-table on TRAIN and VALID:\n" + str(table_train_valid))

        if test_data_loader:
            table_test = create_error_table(
                table_type=args.error_table,
                all_data_loaders=test_data_loader,
                model=model_to_evaluate,
                loss_fn=loss_fn,
                output_args=output_args,
                log_wandb=args.wandb,
                device=device,
                distributed=args.distributed,
            )
            logging.info("Error-table on TEST:\n" + str(table_test))
        if args.plot:
            try:
                plotter = TrainingPlotter(
                    results_dir=logger.path,
                    heads=heads,
                    table_type=args.error_table,
                    train_valid_data=train_valid_data_loader,
                    test_data=test_data_loader,
                    output_args=output_args,
                    device=device,
                    plot_frequency=args.plot_frequency,
                    distributed=args.distributed,
                    swa_start=swa.start if swa else None
                )
                plotter.plot(epoch, model_to_evaluate, rank)
            except Exception as e:  # pylint: disable=W0718
                logging.debug(f"Plotting failed: {e}")

        if args.distributed:
            torch.distributed.barrier()

    logging.info("Done")
    if args.distributed:
        torch.distributed.destroy_process_group()


if __name__ == "__main__":
    main()



###########################################################################################
# Statistics utilities
# Authors: Ilyes Batatia, Gregor Simm, David Kovacs
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import json
import logging
import os
import sys
from pathlib import Path
from typing import Any, Dict, Iterable, Optional, Sequence, Union

import numpy as np
import torch

from .torch_tools import to_numpy


def compute_mae(delta: np.ndarray) -> float:
    return np.mean(np.abs(delta)).item()


def compute_rel_mae(delta: np.ndarray, target_val: np.ndarray) -> float:
    target_norm = np.mean(np.abs(target_val))
    return np.mean(np.abs(delta)).item() / (target_norm + 1e-9) * 100


def compute_rmse(delta: np.ndarray) -> float:
    return np.sqrt(np.mean(np.square(delta))).item()


def compute_rel_rmse(delta: np.ndarray, target_val: np.ndarray) -> float:
    target_norm = np.sqrt(np.mean(np.square(target_val))).item()
    return np.sqrt(np.mean(np.square(delta))).item() / (target_norm + 1e-9) * 100


def compute_q95(delta: np.ndarray) -> float:
    return np.percentile(np.abs(delta), q=95)


def compute_c(delta: np.ndarray, eta: float) -> float:
    return np.mean(np.abs(delta) < eta).item()


def get_tag(name: str, seed: int) -> str:
    return f"{name}_run-{seed}"


def setup_logger(
    level: Union[int, str] = logging.INFO,
    tag: Optional[str] = None,
    directory: Optional[str] = None,
    rank: Optional[int] = 0,
):
    # Create a logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)  # Set to DEBUG to capture all levels

    # Create formatters
    formatter = logging.Formatter(
        "%(asctime)s.%(msecs)03d %(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # Add filter for rank
    logger.addFilter(lambda _: rank == 0)

    # Create console handler
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if directory is not None and tag is not None:
        os.makedirs(name=directory, exist_ok=True)

        # Create file handler for non-debug logs
        main_log_path = os.path.join(directory, f"{tag}.log")
        fh_main = logging.FileHandler(main_log_path)
        fh_main.setLevel(level)
        fh_main.setFormatter(formatter)
        logger.addHandler(fh_main)

        # Create file handler for debug logs
        debug_log_path = os.path.join(directory, f"{tag}_debug.log")
        fh_debug = logging.FileHandler(debug_log_path)
        fh_debug.setLevel(logging.DEBUG)
        fh_debug.setFormatter(formatter)
        fh_debug.addFilter(lambda record: record.levelno >= logging.DEBUG)
        logger.addHandler(fh_debug)


class AtomicNumberTable:
    def __init__(self, zs: Sequence[int]):
        self.zs = zs

    def __len__(self) -> int:
        return len(self.zs)

    def __str__(self):
        return f"AtomicNumberTable: {tuple(s for s in self.zs)}"

    def index_to_z(self, index: int) -> int:
        return self.zs[index]

    def z_to_index(self, atomic_number: str) -> int:
        return self.zs.index(atomic_number)


def get_atomic_number_table_from_zs(zs: Iterable[int]) -> AtomicNumberTable:
    z_set = set()
    for z in zs:
        z_set.add(z)
    return AtomicNumberTable(sorted(list(z_set)))


def atomic_numbers_to_indices(
    atomic_numbers: np.ndarray, z_table: AtomicNumberTable
) -> np.ndarray:
    to_index_fn = np.vectorize(z_table.z_to_index)
    return to_index_fn(atomic_numbers)


class UniversalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, torch.Tensor):
            return to_numpy(o)
        return json.JSONEncoder.default(self, o)


class MetricsLogger:
    def __init__(self, directory: str, tag: str) -> None:
        self.directory = directory
        self.filename = tag + ".txt"
        self.path = os.path.join(self.directory, self.filename)

    def log(self, d: Dict[str, Any]) -> None:
        os.makedirs(name=self.directory, exist_ok=True)
        with open(self.path, mode="a", encoding="utf-8") as f:
            f.write(json.dumps(d, cls=UniversalEncoder))
            f.write("\n")


# pylint: disable=abstract-method, arguments-differ
class LAMMPS_MP(torch.autograd.Function):
    @staticmethod
    def forward(ctx, *args):
        feats, data = args  # unpack
        ctx.vec_len = feats.shape[-1]
        ctx.data = data
        out = torch.empty_like(feats)
        data.forward_exchange(feats, out, ctx.vec_len)
        return out

    @staticmethod
    def backward(ctx, *grad_outputs):
        (grad,) = grad_outputs  # unpack
        gout = torch.empty_like(grad)
        ctx.data.reverse_exchange(grad, gout, ctx.vec_len)
        return gout, None


def get_cache_dir() -> Path:
    # get cache dir from XDG_CACHE_HOME if set, otherwise appropriate default
    return Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache")) / "mace"


def filter_nonzero_weight(
    batch,
    quantity_l,
    weight,
    quantity_weight,
    spread_atoms=False,
    spread_quantity_vector=True,
) -> float:
    quantity = quantity_l[-1]
    # repeat with interleaving for per-atom quantities
    if spread_atoms:
        weight = torch.repeat_interleave(
            weight, batch.ptr[1:] - batch.ptr[:-1]
        ).unsqueeze(-1)
        quantity_weight = torch.repeat_interleave(
            quantity_weight, batch.ptr[1:] - batch.ptr[:-1]
        ).unsqueeze(-1)

    # repeat for additional dimensions
    if len(quantity.shape) > 1:
        repeats = [1] + list(quantity.shape[1:])
        view = [-1] + [1] * (len(quantity.shape) - 1)
        weight = weight.view(*view).repeat(*repeats)
        if spread_quantity_vector:
            quantity_weight = quantity_weight.view(*view).repeat(*repeats)

    filtered_q = quantity[weight * quantity_weight > 0]
    if len(filtered_q) == 0:
        quantity_l.pop()
        return 0.0

    quantity_l[-1] = filtered_q
    return 1.0



import logging
import os
import sys
import time
from contextlib import contextmanager
from typing import Dict, Tuple

import torch
from ase.data import chemical_symbols
from macer.externals.e3nn_legacy.util.jit import compile_mode

try:
    from lammps.mliap.mliap_unified_abc import MLIAPUnified
except ImportError:

    class MLIAPUnified:
        def __init__(self):
            pass


class MACELammpsConfig:
    """Configuration settings for MACE-LAMMPS integration."""

    def __init__(self):
        self.debug_time = self._get_env_bool("MACE_TIME", False)
        self.debug_profile = self._get_env_bool("MACE_PROFILE", False)
        self.profile_start_step = int(os.environ.get("MACE_PROFILE_START", "5"))
        self.profile_end_step = int(os.environ.get("MACE_PROFILE_END", "10"))
        self.allow_cpu = self._get_env_bool("MACE_ALLOW_CPU", False)
        self.force_cpu = self._get_env_bool("MACE_FORCE_CPU", False)

    @staticmethod
    def _get_env_bool(var_name: str, default: bool) -> bool:
        return os.environ.get(var_name, str(default)).lower() in (
            "true",
            "1",
            "t",
            "yes",
        )


@contextmanager
def timer(name: str, enabled: bool = True):
    """Context manager for timing code blocks."""
    if not enabled:
        yield
        return

    start = time.perf_counter()
    try:
        yield
    finally:
        elapsed = time.perf_counter() - start
        logging.info(f"Timer - {name}: {elapsed*1000:.3f} ms")


@compile_mode("script")
class MACEEdgeForcesWrapper(torch.nn.Module):
    """Wrapper that adds per-pair force computation to a MACE model."""

    def __init__(self, model: torch.nn.Module, **kwargs):
        super().__init__()
        self.model = model
        self.register_buffer("atomic_numbers", model.atomic_numbers)
        self.register_buffer("r_max", model.r_max)
        self.register_buffer("num_interactions", model.num_interactions)

        if not hasattr(model, "heads"):
            model.heads = ["Default"]

        head_name = kwargs.get("head", model.heads[-1])
        head_idx = model.heads.index(head_name)
        self.register_buffer("head", torch.tensor([head_idx], dtype=torch.long))

        for p in self.model.parameters():
            p.requires_grad = False

    def forward(
        self, data: Dict[str, torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Compute energies and per-pair forces."""
        data["head"] = self.head

        out = self.model(
            data,
            training=False,
            compute_force=False,
            compute_virials=False,
            compute_stress=False,
            compute_displacement=False,
            compute_edge_forces=True,
            lammps_mliap=True,
        )

        node_energy = out["node_energy"]
        pair_forces = out["edge_forces"]
        total_energy = out["energy"][0]

        if pair_forces is None:
            pair_forces = torch.zeros_like(data["vectors"])

        return total_energy, node_energy, pair_forces


class LAMMPS_MLIAP_MACE(MLIAPUnified):
    """MACE integration for LAMMPS using the MLIAP interface."""

    def __init__(self, model, **kwargs):
        super().__init__()
        self.config = MACELammpsConfig()
        self.model = MACEEdgeForcesWrapper(model, **kwargs)
        self.element_types = [chemical_symbols[s] for s in model.atomic_numbers]
        self.num_species = len(self.element_types)
        self.rcutfac = 0.5 * float(model.r_max)
        self.ndescriptors = 1
        self.nparams = 1
        self.dtype = model.r_max.dtype
        self.device = "cpu"
        self.initialized = False
        self.step = 0

    def _initialize_device(self, data):
        using_kokkos = "kokkos" in data.__class__.__module__.lower()

        if using_kokkos and not self.config.force_cpu:
            device = torch.as_tensor(data.elems).device
            if device.type == "cpu" and not self.config.allow_cpu:
                raise ValueError(
                    "GPU requested but tensor is on CPU. Set MACE_ALLOW_CPU=true to allow CPU computation."
                )
        else:
            device = torch.device("cpu")

        self.device = device
        self.model = self.model.to(device)
        logging.info(f"MACE model initialized on device: {device}")
        self.initialized = True

    def compute_forces(self, data):
        natoms = data.nlocal
        ntotal = data.ntotal
        nghosts = ntotal - natoms
        npairs = data.npairs
        species = torch.as_tensor(data.elems, dtype=torch.int64)

        if not self.initialized:
            self._initialize_device(data)

        self.step += 1
        self._manage_profiling()

        if natoms == 0 or npairs <= 1:
            return

        with timer("total_step", enabled=self.config.debug_time):
            with timer("prepare_batch", enabled=self.config.debug_time):
                batch = self._prepare_batch(data, natoms, nghosts, species)

            with timer("model_forward", enabled=self.config.debug_time):
                _, atom_energies, pair_forces = self.model(batch)

                if self.device.type != "cpu":
                    torch.cuda.synchronize()

            with timer("update_lammps", enabled=self.config.debug_time):
                self._update_lammps_data(data, atom_energies, pair_forces, natoms)

    def _prepare_batch(self, data, natoms, nghosts, species):
        """Prepare the input batch for the MACE model."""
        return {
            "vectors": torch.as_tensor(data.rij).to(self.dtype).to(self.device),
            "node_attrs": torch.nn.functional.one_hot(
                species.to(self.device), num_classes=self.num_species
            ).to(self.dtype),
            "edge_index": torch.stack(
                [
                    torch.as_tensor(data.pair_j, dtype=torch.int64).to(self.device),
                    torch.as_tensor(data.pair_i, dtype=torch.int64).to(self.device),
                ],
                dim=0,
            ),
            "batch": torch.zeros(natoms, dtype=torch.int64, device=self.device),
            "lammps_class": data,
            "natoms": (natoms, nghosts),
        }

    def _update_lammps_data(self, data, atom_energies, pair_forces, natoms):
        """Update LAMMPS data structures with computed energies and forces."""
        if self.dtype == torch.float32:
            pair_forces = pair_forces.double()
        eatoms = torch.as_tensor(data.eatoms)
        eatoms.copy_(atom_energies[:natoms])
        data.energy = torch.sum(atom_energies[:natoms])
        data.update_pair_forces_gpu(pair_forces)

    def _manage_profiling(self):
        if not self.config.debug_profile:
            return

        if self.step == self.config.profile_start_step:
            logging.info(f"Starting CUDA profiler at step {self.step}")
            torch.cuda.profiler.start()

        if self.step == self.config.profile_end_step:
            logging.info(f"Stopping CUDA profiler at step {self.step}")
            torch.cuda.profiler.stop()
            logging.info("Profiling complete. Exiting.")
            sys.exit()

    def compute_descriptors(self, data):
        pass

    def compute_gradients(self, data):
        pass



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import subprocess
import sys


def run(cmd: list[str]) -> subprocess.CompletedProcess:
    return subprocess.run(cmd, text=True, capture_output=True, check=False)


def require(cond: bool, msg: str, failures: list[str]):
    if not cond:
        failures.append(msg)


def main() -> int:
    failures: list[str] = []

    # 1) Command exposure
    help_root = run([sys.executable, "-m", "macer.cli.phonopy_main", "-h"])
    require(help_root.returncode == 0, "root help command failed", failures)
    root_text = help_root.stdout + help_root.stderr
    require("qscaild" in root_text, "qscaild command is not exposed in CLI help", failures)

    # 2) qscaild parser options
    help_qs = run([sys.executable, "-m", "macer.cli.phonopy_main", "qscaild", "-h"])
    require(help_qs.returncode == 0, "qscaild -h failed", failures)
    qs_text = help_qs.stdout + help_qs.stderr
    for opt in [
        "--nconf",
        "--nfits",
        "--memory",
        "--mixing",
        "--tolerance",
        "--use-pressure",
        "--pressure-diag",
        "--dry-run",
    ]:
        require(opt in qs_text, f"missing qscaild option in help: {opt}", failures)

    # 3) Parser defaults via dry-run (no expensive calculation)
    dry = run(
        [
            sys.executable,
            "-m",
            "macer.cli.phonopy_main",
            "qscaild",
            "-p",
            "README.md",  # intentionally invalid structure file for a parse-only check
            "-T",
            "300",
            "--dry-run",
        ]
    )
    # Expect non-zero because README.md is not a valid structure, but parser+dispatch must run.
    out = (dry.stdout or "") + (dry.stderr or "")
    require(
        ("Dry-run enabled." in out) or ("not a valid POSCAR/CIF" in out) or ("Error" in out),
        "qscaild dispatch did not execute as expected in dry-run path",
        failures,
    )

    if failures:
        print("qSCAILD validation: FAILED")
        for item in failures:
            print(f"- {item}")
        return 1

    print("qSCAILD validation: PASSED")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())




from typing import Optional
import torch
import numpy as np
from scipy.spatial import KDTree

def radius_graph(
    x: torch.Tensor,
    r: float,
    batch: Optional[torch.Tensor] = None,
    loop: bool = False,
    max_num_neighbors: int = 32,
    flow: str = 'source_to_target',
    num_workers: int = 1,
    batch_size: Optional[int] = None,
) -> torch.Tensor:
    """
    Fallback implementation of torch_cluster.radius_graph using scipy.spatial.KDTree.
    This is used when torch_cluster is not installed (e.g., on some HPC systems).
    """
    try:
        import torch_cluster
        return torch_cluster.radius_graph(
            x=x, r=r, batch=batch, loop=loop, 
            max_num_neighbors=max_num_neighbors, 
            flow=flow, num_workers=num_workers, batch_size=batch_size
        )
    except ImportError:
        pass

    device = x.device
    x_np = x.detach().cpu().numpy()
    
    if batch is None:
        batch = torch.zeros(x.size(0), dtype=torch.long, device=device)
    
    batch_np = batch.detach().cpu().numpy()
    
    row, col = [], []
    for b in np.unique(batch_np):
        idx = np.where(batch_np == b)[0]
        if len(idx) == 0:
            continue
        
        # Build KDTree for the current batch
        tree = KDTree(x_np[idx])
        
        # query_ball_tree returns indices within the current batch slice
        indices = tree.query_ball_tree(tree, r)
        
        for i, neighbors in enumerate(indices):
            count = 0
            for j in neighbors:
                if not loop and i == j:
                    continue
                if count >= max_num_neighbors:
                    break
                
                # row is source, col is target
                if flow == 'source_to_target':
                    row.append(idx[j])
                    col.append(idx[i])
                else:
                    row.append(idx[i])
                    col.append(idx[j])
                count += 1
                
    if not row:
        return torch.empty((2, 0), dtype=torch.long, device=device)
        
    return torch.tensor([row, col], dtype=torch.long, device=device)


def radius(
    x: torch.Tensor,
    y: torch.Tensor,
    r: float,
    batch_x: Optional[torch.Tensor] = None,
    batch_y: Optional[torch.Tensor] = None,
    max_num_neighbors: int = 32,
    num_workers: int = 1,
) -> torch.Tensor:
    """
    Fallback implementation of torch_cluster.radius using scipy.spatial.KDTree.
    """
    try:
        import torch_cluster
        return torch_cluster.radius(
            x=x, y=y, r=r, batch_x=batch_x, batch_y=batch_y, 
            max_num_neighbors=max_num_neighbors, num_workers=num_workers
        )
    except ImportError:
        pass

    device = x.device
    x_np = x.detach().cpu().numpy()
    y_np = y.detach().cpu().numpy()
    
    if batch_x is None:
        batch_x = torch.zeros(x.size(0), dtype=torch.long, device=device)
    if batch_y is None:
        batch_y = torch.zeros(y.size(0), dtype=torch.long, device=device)
        
    batch_x_np = batch_x.detach().cpu().numpy()
    batch_y_np = batch_y.detach().cpu().numpy()
    
    row, col = [], []
    for b in np.unique(batch_y_np):
        idx_x = np.where(batch_x_np == b)[0]
        idx_y = np.where(batch_y_np == b)[0]
        
        if len(idx_x) == 0 or len(idx_y) == 0:
            continue
            
        tree_x = KDTree(x_np[idx_x])
        tree_y = KDTree(y_np[idx_y])
        
        # Find neighbors of y in x
        indices = tree_y.query_ball_tree(tree_x, r)
        
        for i, neighbors in enumerate(indices):
            count = 0
            for j in neighbors:
                if count >= max_num_neighbors:
                    break
                
                # row is index into x, col is index into y
                row.append(idx_x[j])
                col.append(idx_y[i])
                count += 1
                
    if not row:
        return torch.empty((2, 0), dtype=torch.long, device=device)
        
    return torch.tensor([row, col], dtype=torch.long, device=device)



# This coded was modified from the cuequivariance library: https://github.com/NVIDIA/cuEquivariance/blob/main/cuequivariance/cuequivariance/group_theory/experimental/mace/symmetric_contractions.py
# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Optional

import numpy as np
from macer.externals.e3nn_legacy import o3

from macer.externals.mace_bundled.tools.cg import U_matrix_real

try:
    import cuequivariance as cue
    from cuequivariance.etc.linalg import round_to_sqrt_rational, triu_array

    CUEQQ_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    CUEQQ_AVAILABLE = False

    class DummyCueq:
        class EquivariantPolynomial:
            pass

        class Irreps:
            pass

    cue = DummyCueq()
    round_to_sqrt_rational = None
    triu_array = None


def symmetric_contraction_proj(
    irreps_in: cue.Irreps, irreps_out: cue.Irreps, degrees: tuple[int, ...]
) -> tuple[cue.EquivariantPolynomial, np.ndarray]:
    r"""
    subscripts: ``weights[u],input[u],output[u]``

    Example:

    .. code-block:: python

        e, p = symmetric_contraction(
            4 * cue.Irreps("SO3", "0+1+2"), 4 * cue.Irreps("SO3", "0+1"), [1, 2, 3]
        )
        assert p.shape == (62, 18)

        mul = e.inputs[1].irreps.muls[0]
        w = jax.random.normal(jax.random.key(0), (p.shape[0], mul))
        w = jnp.einsum("au,ab->bu", w, p).flatten()

        x = cuex.randn(jax.random.key(1), e.inputs[1])
        y = cuex.equivariant_polynomial(e, [w, x])
    """
    return symmetric_contraction_cached(irreps_in, irreps_out, tuple(degrees))


def symmetric_contraction_cached(
    irreps_in: cue.Irreps, irreps_out: cue.Irreps, degrees: tuple[int, ...]
) -> tuple[cue.EquivariantPolynomial, np.ndarray]:
    assert min(degrees) > 0

    # poly1 replicates the behavior of the original MACE implementation
    poly1 = cue.EquivariantPolynomial.stack(
        [
            cue.EquivariantPolynomial.stack(
                [
                    _symmetric_contraction(irreps_in, irreps_out[i : i + 1], deg)
                    for deg in reversed(degrees)
                ],
                [True, False, False],
            )
            for i in range(len(irreps_out))
        ],
        [True, False, True],
    )

    poly2 = cue.descriptors.symmetric_contraction(irreps_in, irreps_out, degrees)
    a1, a2 = [
        np.concatenate(
            [
                _flatten(
                    _stp_to_matrix(d.symmetrize_operands(range(1, d.num_operands - 1))),
                    1,
                    None,
                )
                for _, d in pol.polynomial.operations
            ],
            axis=1,
        )
        for pol in [poly1, poly2]
    ]

    # This nonzeros selection is just for lightening the inversion
    nonzeros = np.nonzero(np.any(a1 != 0, axis=0) | np.any(a2 != 0, axis=0))[0]
    a1, a2 = a1[:, nonzeros], a2[:, nonzeros]
    projection_1_2 = a1 @ np.linalg.pinv(a2)
    # projection = np.linalg.lstsq(a2.T, a1.T, rcond=None)[0].T
    projection_1_2 = round_to_sqrt_rational(projection_1_2)
    np.testing.assert_allclose(a1, projection_1_2 @ a2, atol=1e-7)
    return poly2, projection_1_2


def _flatten(
    x: np.ndarray, axis_start: Optional[int] = None, axis_end: Optional[int] = None
) -> np.ndarray:
    x = np.asarray(x)
    if axis_start is None:
        axis_start = 0
    if axis_end is None:
        axis_end = x.ndim
    assert 0 <= axis_start <= axis_end <= x.ndim
    return x.reshape(
        x.shape[:axis_start]
        + (np.prod(x.shape[axis_start:axis_end]),)
        + x.shape[axis_end:]
    )


def _stp_to_matrix(
    d: cue.SegmentedTensorProduct,
) -> np.ndarray:
    m = np.zeros([ope.num_segments for ope in d.operands])
    for path in d.paths:
        m[path.indices] = path.coefficients
    return m


# This function is an adaptation of https://github.com/ACEsuit/mace/blob/bd412319b11c5f56c37cec6c4cfae74b2a49ff43/mace/modules/symmetric_contraction.py
def _symmetric_contraction(
    irreps_in: cue.Irreps, irreps_out: cue.Irreps, degree: int
) -> cue.EquivariantPolynomial:
    mul = irreps_in.muls[0]
    assert all(mul == m for m in irreps_in.muls)
    assert all(mul == m for m in irreps_out.muls)
    irreps_in = irreps_in.set_mul(1)
    irreps_out = irreps_out.set_mul(1)

    input_operands = range(1, degree + 1)
    output_operand = degree + 1

    abc = "abcdefgh"[:degree]
    d = cue.SegmentedTensorProduct.from_subscripts(
        f"u_{'_'.join(f'{a}' for a in abc)}_i+{abc}ui"
    )

    for i in input_operands:
        d.add_segment(i, (irreps_in.dim,))
    irreps_out_e3nn = o3.Irreps(str(irreps_out))
    irreps_in_e3nn = o3.Irreps(str(irreps_in))
    for _, ir in irreps_out:
        u = U_matrix_real(irreps_in_e3nn, irreps_out_e3nn, degree)[-1]
        if str(ir) == "0e" or str(ir) == "0o":
            u = u.unsqueeze(0)
        u = np.asarray(u)
        u = np.moveaxis(u, 0, -1)
        # u is shape (irreps_in.dim, ..., irreps_in.dim, u, ir_out.dim)

        if u.shape[-2] == 0:
            d.add_segment(output_operand, {"i": ir.dim})
        else:
            u = triu_array(u, degree)
            d.add_path(None, *(0,) * degree, None, c=u)

    d = d.flatten_coefficient_modes()
    d = d.append_modes_to_all_operands("u", {"u": mul})

    assert d.num_operands >= 3
    [w, x], y = d.operands[:2], d.operands[-1]

    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps_in.new_scalars(w.size), cue.ir_mul),
            cue.IrrepsAndLayout(mul * irreps_in, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(mul * irreps_out, cue.ir_mul)],
        cue.SegmentedPolynomial(
            [w, x], [y], [(cue.Operation([0] + [1] * degree + [2]), d)]
        ),
    )



"""
Macer: Machine-learning accelerated Atomic Computational Environment for automated Research workflows
Copyright (c) 2025 The Macer Package Authors
Author: Soungmin Bae <soungminbae@gmail.com>
License: MIT
"""

import sys
import os
import shutil
from argparse import Namespace
from pathlib import Path
import yaml

from pymatgen.core import Structure, Composition
from vise.cli.main_functions import get_poscar_from_mp
from pydefect.cli.vasp.main_vasp_functions import make_competing_phase_dirs
from pydefect.cli.main_functions import make_standard_and_relative_energies
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag, RelativeEnergies, TargetVertices, ChemPotDiagMaker
from pydefect.defaults import defaults as pydefect_defaults

from macer.pydefect.utils import run_macer_relax, stabilize_target, generate_composition_energies_direct, maybe_plot_cpd_pdf
from macer.utils.validation import check_poscar_format

def run_cpd_workflow(args):
    """
    Auto CPD workflow.
    args requires: formula, mpid, doping. Optional: poscar
    """
    
    # ------------------------------------------------------------------
    # Step 1: Get POSCAR and Determine Elements
    # ------------------------------------------------------------------
    target_formula = None
    host_elements = []
    competing_elements = []
    
    # If POSCAR is provided, use it to determine formula and elements
    if args.poscar:
        from macer.utils.struct_tools import resolve_structure_inputs
        resolved = resolve_structure_inputs(poscar_paths=[args.poscar])
        if not resolved:
            print("Error: No valid POSCAR provided.")
            return
        args.poscar = resolved[0]
        print(f"--- Step 1: Read input POSCAR from {args.poscar} ---")
        input_poscar_path = Path(args.poscar).absolute()
        if not input_poscar_path.exists():
            print(f"Error: {input_poscar_path} does not exist.")
            return
        
        # Validate POSCAR format (VASP 5 check)
        try:
            check_poscar_format(input_poscar_path)
        except ValueError as e:
            print(f"\nError: {e}\n")
            return

        try:
            structure = Structure.from_file(str(input_poscar_path))
            host_elements = [str(e) for e in structure.composition.elements]
            target_formula = structure.composition.reduced_formula
            print(f"Host Elements: {host_elements}, Target Formula: {target_formula}")
            
        except Exception as e:
            print(f"Failed to read input POSCAR: {e}")
            return

    else:
        # Check if mpid or formula is provided to download (Prioritize mpid)
        if args.mpid:
            target_str = args.mpid
            print(f"--- Step 1: Retrieve POSCAR for {target_str} (mpid) ---")
            args_gp = Namespace(mpid=args.mpid, formula=None)
        elif args.formula:
            target_str = args.formula
            print(f"--- Step 1: Retrieve POSCAR for {target_str} (formula) ---")
            args_gp = Namespace(mpid=None, formula=args.formula)
        else:
            print("\nError: No structure specified. Provide -p, -f or -m.")
            return

        try:
            from macer.utils.struct_tools import _download_poscar_from_mp
            _download_poscar_from_mp(mpid=args.mpid, formula=args.formula, workdir=".", silent=True)
            if Path("POSCAR").exists():
                print(f"Successfully retrieved POSCAR for {target_str}.")
            else:
                print("Error: POSCAR file was not created.")
                return
        except Exception as e:
            print("Error: Failed to retrieve POSCAR from Materials Project.")
            print(f"Details: {e}")
            print("Please verify your MP API settings.")
            print("You must set an API key, e.g.:")
            print('  export MP_API_KEY=\"YOUR_API_KEY\"')
            print("Get an API key from Materials Project and add it to your shell rc (e.g., ~/.bashrc or ~/.zshrc).")
            return

        # Determine elements from the retrieved POSCAR
        try:
            structure = Structure.from_file("POSCAR")
            host_elements = [str(e) for e in structure.composition.elements]
            target_formula = structure.composition.reduced_formula
            print(f"Host Elements: {host_elements}, Target Formula: {target_formula}")
        except Exception as e:
            print(f"Failed to read POSCAR or determine elements: {e}")
            return

    # Determine competing elements
    competing_elements = host_elements.copy()
    if args.doping:
        for dopant in args.doping:
            if dopant not in competing_elements:
                competing_elements.append(dopant)
        print(f"Dopant(s) added: {args.doping}. Full competing element list: {competing_elements}")

    # ------------------------------------------------------------------
    # Create CPD Output Directory and Move Files
    # ------------------------------------------------------------------
    dopants_part = ""
    if args.doping:
        dopants_part = f"-DOPANG={'_'.join(args.doping)}"
    
    if getattr(args, "output_dir", None):
        cpd_output_dir = Path(args.output_dir)
    else:
        base_dir_name = f"CPD-target={target_formula}{dopants_part}-mlff={args.ff}"
        # Determine base path for output directory
        base_path = Path.cwd()
        if args.poscar:
            base_path = Path(args.poscar).absolute().parent
        cpd_output_dir = base_path / base_dir_name
    
    i = 1
    original_base = cpd_output_dir
    while cpd_output_dir.exists():
        cpd_output_dir = Path(f"{original_base}-NEW{i:03d}")
        i += 1
    
    print(f"\nCreating output directory: {cpd_output_dir.name}")
    print(f"Location: {cpd_output_dir.parent}")
    cpd_output_dir.mkdir()
    
    # Move/Copy initial files to the new directory
    if args.poscar:
        # Copy user provided POSCAR to the workspace
        shutil.copy(str(Path(args.poscar).absolute()), cpd_output_dir / "POSCAR")
    else:
        # Move downloaded files
        for filename in ["POSCAR", "prior_info.yaml"]:
            if Path(filename).exists():
                shutil.move(filename, cpd_output_dir / filename)
            
    # Change working directory
    original_cwd = Path.cwd()
    os.chdir(cpd_output_dir)
    print(f"Changed working directory to: {os.getcwd()}")

    try:
        # ------------------------------------------------------------------
        # Step 2: Generate competing phase directories
        # ------------------------------------------------------------------
        print(f"\n--- Step 2: Generate competing phase directories ({'-'.join(competing_elements)}) ---")
        args_mp = Namespace(elements=competing_elements, 
                            e_above_hull=pydefect_defaults.e_above_hull)
        try:
            make_competing_phase_dirs(args_mp)
            print("Competing phase directories created.")
        except Exception as e:
            print(f"Failed to generate competing phase directories: {e}")
            return

        # ------------------------------------------------------------------
        # Step 3: Run macer relax in each directory
        # ------------------------------------------------------------------
        print("\n--- Step 3: Run macer relax (Single MLFF Load) ---")
        
        # If input POSCAR was provided, ensure it is included as {formula}_host
        host_dir_name = f"{target_formula}_host"
        if args.poscar:
             host_dir = Path(host_dir_name)
             if not host_dir.exists():
                 host_dir.mkdir()
                 shutil.copy("POSCAR", host_dir / "POSCAR")
                 # Create dummy prior_info.yaml if not exists, as some tools might check for it
                 if not (host_dir / "prior_info.yaml").exists():
                      # Minimal content
                      with open(host_dir / "prior_info.yaml", "w") as f:
                          yaml.dump({"band_gap": 0.0, "total_magnetization": 0.0}, f)
                 print(f"Created {host_dir_name} from input POSCAR.")
        
        target_dirs = []
        for d in Path.cwd().iterdir():
            if d.is_dir() and (d / "POSCAR").exists():
                 # Check for prior_info.yaml OR if it is the host dir
                 if (d / "prior_info.yaml").exists() or d.name == host_dir_name:
                     target_dirs.append(d)
        
        target_dirs.sort(key=lambda x: x.name)
        
        if not target_dirs:
            print("No defect directories found to relax.")
            return

        # Use the utility function for relaxation (ISIF=3 for CPD)
        successful_dirs = run_macer_relax(
            target_dirs, 
            isif=3, 
            verbose=True, 
            fmax=args.fmax,
            ff=args.ff,
            model_path=args.model,
            device=args.device,
            modal=args.modal,
            n_jobs=getattr(args, "n_jobs", 1),
        )

        # ------------------------------------------------------------------
        # Step 4: Generate composition_energies.yaml
        # ------------------------------------------------------------------
        print("\n--- Step 4: Generate composition_energies.yaml ---")
        if successful_dirs:
            yaml_filename = "composition_energies.yaml"
            try:
                generate_composition_energies_direct(successful_dirs, yaml_filename)
            except Exception as e:
                print(f"Failed to create composition_energies.yaml: {e}")
        else:
            print("No successful calculations to process.")

        # ------------------------------------------------------------------
        # Step 5: Generate Standard and Relative Energies
        # ------------------------------------------------------------------
        print("\n--- Step 5: Generate Standard and Relative Energies ---")
        if Path("composition_energies.yaml").exists():
            # Filter composition_energies to exclude cross-impurity compounds
            try:
                with open("composition_energies.yaml", 'r') as f:
                    comp_energies = yaml.safe_load(f)
                
                filtered_comp_energies = {}
                removed_compounds = []
                
                for formula, data in comp_energies.items():
                    comp = Composition(formula)
                    elements_in_comp = [str(e) for e in comp.elements]
                    impurities_in_comp = [e for e in elements_in_comp if e not in host_elements]
                    
                    if len(impurities_in_comp) <= 1:
                        filtered_comp_energies[formula] = data
                    else:
                        removed_compounds.append(formula)
                
                if removed_compounds:
                    print(f"Filtering out compounds with multiple impurities: {removed_compounds}")
                
                filtered_yaml_filename = "composition_energies_filtered.yaml"
                with open(filtered_yaml_filename, 'w') as f:
                    yaml.dump(filtered_comp_energies, f)
                    
                args_sre = Namespace(composition_energies_yaml=filtered_yaml_filename)
                make_standard_and_relative_energies(args_sre)
                print("Generated standard_energies.yaml and relative_energies.yaml (from filtered data).")
                
            except Exception as e:
                print(f"Failed to generate standard/relative energies: {e}")
        else:
            print("composition_energies.yaml not found. Skipping Step 5.")

        # ------------------------------------------------------------------
        # Step 6: Chemical Potential Diagram
        # ------------------------------------------------------------------
        print(f"\n--- Step 6: Generate Chemical Potential Diagram for {target_formula} ---")
        if Path("relative_energies.yaml").exists() and target_formula:
            
            # Check for stability and apply shift if necessary
            applied_shift = stabilize_target("relative_energies.yaml", target_formula, manual_shift=args.energy_shift_target)
            
            # Print Convex Hull Summary
            if Path("convex_hull_energy.yaml").exists():
                with open("convex_hull_energy.yaml", 'r') as f:
                    ch_info = yaml.safe_load(f)
                
                print(f"\n--- Convex Hull Stability Summary for {target_formula} ---")
                print(f"  Target Formation Energy: {ch_info['formation_energy_per_atom']:.4f} eV/atom")
                print(f"  Convex Hull Energy     : {ch_info['convex_hull_energy_per_atom']:.4f} eV/atom")
                print(f"  Energy Above Hull      : {ch_info['e_above_hull']:.4f} eV/atom")
                
                decomp_info = ch_info.get('decomposition', {})
                dist = decomp_info.get('hull_distance', 0.0)
                status = "STABLE" if dist < 0 else "UNSTABLE"
                print(f"  Hull Distance (Decomp.): {dist:.4f} eV/atom ({status})")
                
                phases = decomp_info.get('competing_phases', {})
                decomp_str = " + ".join([f"{amt:.3f} {phi}" for phi, amt in phases.items()])
                print(f"  Decomposition Products : {decomp_str}")
            
            # Note: We use host_elements for the vertices of the CPD
            try:
                # Check for unary system to avoid Qhull error
                if len(host_elements) == 1:
                    print(f"Unary system detected ({host_elements[0]}). Manually generating target_vertices.yaml.")
                    manual_vertices = {
                        "target": target_formula,
                        "A": {
                            "chem_pot": {host_elements[0]: 0.0},
                            "competing_phases": [],
                            "impurity_phases": []
                        }
                    }
                    with open("target_vertices.yaml", "w") as f:
                        yaml.dump(manual_vertices, f)
                    print("Generated target_vertices.yaml manually.")
                else:
                    # Direct API usage instead of make_cpd_and_vertices(args_cv)
                    rel_energies = RelativeEnergies.from_yaml("relative_energies.yaml")
                    cpd_maker = ChemPotDiagMaker(rel_energies, elements=host_elements, target=target_formula)
                    cpd = cpd_maker.chem_pot_diag
                    cpd.to_json_file()
                    
                    target_vertices = cpd.to_target_vertices
                    target_vertices.to_yaml_file("target_vertices.yaml")
                    
                    print("Generated chem_pot_diag.json and target_vertices.yaml.")

                if Path("chem_pot_diag.json").exists():
                    maybe_plot_cpd_pdf("chem_pot_diag.json")
                else:
                    print("chem_pot_diag.json not found. Skipping cpd.pdf.")

                if Path("chem_pot_diag.json").exists():
                    maybe_plot_cpd_pdf("chem_pot_diag.json")
                else:
                    print("chem_pot_diag.json not found. Skipping cpd.pdf.")
                
                if applied_shift != 0.0 and Path("target_vertices.yaml").exists():
                    with open("target_vertices.yaml", "a") as f:
                        f.write(f"# Energy shift applied to make target stable: {applied_shift} eV/atom\n")
                    print(f"Appended stability shift info to target_vertices.yaml")
                    
            except Exception as e:
                print(f"Failed to generate chemical potential diagram: {e}")
        else:
            print("relative_energies.yaml not found or target formula missing. Skipping Step 6.")

    finally:
        # os.chdir(original_cwd) # Not strictly necessary if script ends here, but good practice
        pass
